from __future__ import annotations
from dataclasses import dataclass
import typing_extensions as t
from . import build, parser


@dataclass
class Template:
    _func: t.Callable[..., str]
    _parse_info: parser.Template

    def render(self, **context: t.Any) -> str:
        return self._func(**context)


class Environment:
    _templates: t.Dict[str, Template]
    _globals: t.Dict[str, Template]
    _from_string_cache: t.Dict[str, Template]

    def __init__(
        self,
        templates: t.Optional[t.Dict[str, Template]] = None,
        globals: t.Optional[t.Dict[str, t.Any]] = None,
    ):
        self._templates = templates or {}
        self._globals = globals or {}
        self._from_string_cache = {}

    def get_template(self, name: str) -> t.Optional[Template]:
        return self._templates.get(name, None)

    def render_template(self, name: str, **context: t.Any) -> str:
        template = self.get_template(name)
        if template is None:
            raise ValueError(f"Template {name} not found")

        return template.render(**context)

    def from_string(self, template: str) -> Template:
        if template in self._from_string_cache:
            return self._from_string_cache[template]

        parsed_template = parser.parse_template("<string>", template)
        templates = [template._parse_info for template in self._templates.values()]
        func = build.build_single_template(
            parsed_template,
            templates,
            self._globals,
        )
        template_obj = Template(
            _func=func,
            _parse_info=parsed_template,
        )
        self._from_string_cache[template] = template_obj
        return template_obj



import ast
import builtins
from functools import lru_cache
import typing_extensions as t
from .. import ast_utils
from .constants import (
    KWARGS_VARIABLE,
    NAME_LOOKUP_VARIABLE,
    WITH_STYLES_PARAMETER,
    create_resolve_call,
)


def create_resolve_for_unknown_variables(
    body: t.List[ast.stmt], known_names: t.List[str]
):
    transformer = NameTransformer(known_names)
    for node in body:
        transformer.visit(node)


class NameTransformer(ast.NodeTransformer):
    RESERVED_NAMES = (
        KWARGS_VARIABLE,
        WITH_STYLES_PARAMETER,
        NAME_LOOKUP_VARIABLE,
    )
    known_names: t.List[str]

    def __init__(self, known_names: t.List[str]):
        self.known_names = known_names

    def visit_Assign(self, node: ast.Assign):
        for target in node.targets:
            if isinstance(target, ast.Name):
                self.known_names.append(target.id)

        return self.generic_visit(node)

    def visit_Name(self, node: ast.Name):
        return self.resolve(node)

    def visit_For(self, node: ast.For):
        loop_vars = extract_loop_variables(node.target)
        self.known_names.extend(loop_vars)
        output_node = self.generic_visit(node)
        self.known_names = self.known_names[: -len(loop_vars)]
        return output_node

    def visit_ListComp(self, node: ast.ListComp):
        return self._comprehension(node)

    def visit_SetComp(self, node: ast.SetComp):
        return self._comprehension(node)

    def visit_DictComp(self, node: ast.DictComp):
        return self._comprehension(node)

    def visit_GeneratorExp(self, node: ast.GeneratorExp):
        return self._comprehension(node)

    def _comprehension(
        self, node: t.Union[ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp]
    ):
        loop_vars = []
        for generator in node.generators:
            loop_vars.extend(extract_loop_variables(generator.target))

        self.known_names.extend(loop_vars)

        if isinstance(node, ast.DictComp):
            node.key = self.visit(node.key)
            node.value = self.visit(node.value)
        else:
            node.elt = self.visit(node.elt)

        node.generators = [
            t.cast(ast.comprehension, self.generic_visit(generator))
            for generator in node.generators
        ]

        for _ in range(len(loop_vars)):
            self.known_names.pop()

        return node

    def resolve(self, name: ast.Name) -> ast.expr:
        if (
            name.id.startswith("__")
            or name.id in self.RESERVED_NAMES
            or name.id in self.known_names
            or is_builtin(name.id)
        ):
            return name

        return create_resolve_call(name.id)


def extract_loop_variables(target: ast.expr) -> t.List[str]:
    if isinstance(target, ast.Name):
        # `for a in x:`
        return [target.id]
    elif isinstance(target, ast.Tuple):
        # `for a, b in x:`
        return [elt.id for elt in target.elts if isinstance(elt, ast.Name)]
    else:
        # ¯\_(ツ)_/¯
        # TODO: Deal with case for loop is non standard
        return []


@lru_cache(maxsize=2048)
def is_builtin(name: str):
    if name in dir(builtins):
        return True

    try:
        eval(name, {}, {})
    except NameError:
        return False
    else:
        return True



import ast
import string
import typing_extensions as t
from .. import ast_utils

CSS_VARIABLE = "__css"
WITH_STYLES_PARAMETER = "with_styles"
OUTPUT_VARIABLE = "__html"
NAME_LOOKUP_VARIABLE = "_name_lookup"
KWARGS_VARIABLE = "context"


FILE_HEADER = ast_utils.parse(
    f"""
from __future__ import annotations as _
from tempered._internals import escape as __escape
import typing_extensions as t

__globals = {{}}

def __register_global(name: str, value: t.Any):
    __globals[name] = value

def __resolve(name: str, context: dict[str, t.Any]) -> t.Any:
    if name in context:
        return context[name]
    elif name in __globals:
        return __globals[name]
    else:
        raise RuntimeError(f"The variable '{{name}}' could not be resolved")
"""
)


def create_escape_call(value: ast.expr) -> ast.expr:
    return ast_utils.Call(func=ast_utils.Name("__escape"), arguments=[value])


def component_func_name(template_name: str) -> str:
    return "__" + filter_non_ident_chars(template_name)


def layout_func_name(template_name: str) -> str:
    template_name = filter_non_ident_chars(template_name)
    return f"__{template_name}_layout"


def slot_variable_name(slot_name: t.Union[str, None]) -> str:
    if slot_name is None:
        return OUTPUT_VARIABLE
    else:
        return f"__{slot_name}_slot"


def slot_parameter(slot_name: t.Union[str, None]) -> str:
    if slot_name:
        return f"__slot_{slot_name}"
    else:
        return "__slot_default"


def filter_non_ident_chars(name: str) -> str:
    IDENT_CHARS = string.ascii_letters + string.digits + "_"
    name = "".join(char for char in name if char in IDENT_CHARS)
    return name


def create_resolve_call(name: str):
    return ast_utils.Call(
        func=ast_utils.Name("__resolve"),
        arguments=[
            ast_utils.Constant(name),
            ast_utils.Name(KWARGS_VARIABLE),
        ],
    )


def create_layout_call(
    layout_name: str,
    default_slot: ast.expr,
    css: ast.expr,
    has_default_slot: bool,
    blocks: t.Set[str],
) -> ast.expr:
    kw_args: t.Dict[str, ast.expr] = {}
    kw_args[CSS_VARIABLE] = css
    kw_args[WITH_STYLES_PARAMETER] = ast_utils.Name(WITH_STYLES_PARAMETER)

    if has_default_slot:
        kw_args[slot_parameter(None)] = default_slot

    for slot in blocks:
        kw_args[slot_parameter(slot)] = ast_utils.Name(slot_variable_name(slot))

    return ast_utils.Call(
        func=ast_utils.Name(layout_func_name(layout_name)),
        keywords=kw_args,
        kwargs=ast_utils.Name(KWARGS_VARIABLE),
    )



import sys
from array import array
from pathlib import Path
import typing_extensions as t
from ..errors import ParserException

TToken = t.TypeVar("TToken")


class Scanner(t.Generic[TToken]):
    stream: t.Sequence[TToken]

    def __init__(self, stream: t.Sequence[TToken]):
        self.stream = stream

    @property
    def has_tokens(self) -> bool:
        return len(self.stream) != 0

    def pop(self) -> TToken:
        popped = self.stream[0]
        self.stream = self.stream[1:]
        return popped

    def is_next(self, *token: t.Type[TToken]):
        if self.has_tokens and isinstance(self.stream[0], token):
            return True
        else:
            return False

    T = t.TypeVar("T")

    def accept(self, token: t.Type[T]) -> t.Optional[T]:
        if self.is_next(token):  # type: ignore
            return self.pop()  # type: ignore
        else:
            return None

    def expect(self, token: t.Type[T]) -> T:
        if not self.is_next(token):  # type: ignore
            raise ValueError(f"Expected {token} but got {self.stream[0]!r}")
        else:
            return self.pop()  # type: ignore


class TextScanner:
    original: str
    position: int = 0
    file: t.Union[Path, None]

    if t.TYPE_CHECKING:
        text: array[str]

    def __init__(self, html: str, file: t.Union[Path, None] = None):
        self.file = file
        self.original = html
        if sys.version_info >= (3, 13):
            self.text = array("w", html[::-1])
        else:
            self.text = array("u", html[::-1])

    @property
    def has_text(self) -> bool:
        return len(self.text) != 0

    def pop(self) -> str:
        return self.text.pop()

    def accept(self, text: str) -> bool:
        if not self.startswith(text):
            return False

        self.text = self.text[: -len(text)]
        self.position += len(text)
        return True

    def expect(self, *text: str):
        match = ""
        for match in text:
            if self.accept(match):
                return

        raise self.error(f"Expected {match!r}")

    def startswith(self, *text: str) -> bool:
        for match in text:
            prefix = self.text[-len(match) : :][::-1].tounicode()

            if prefix == match:
                return True

        return False

    def pop_many(self, length: int) -> str:
        if len(self.text) < length:
            self.error("Unexpected end of text")

        popped_text = ""
        for _ in range(length):
            popped_text += self.pop()

        self.position += length
        return popped_text

    def take_optional(self, *text: str) -> t.Union[str, None]:
        for match in text:
            if not self.startswith(match):
                continue

            self.expect(match)
            return match

        return None

    def take(self, *text: str) -> str:
        value = self.take_optional(*text)
        if value is None:
            raise self.error(f"Expected one of {','.join(text)}")

        return value

    def take_until(self, matches: t.Union[str, t.Sequence[str]]) -> str:
        if isinstance(matches, str):
            matches = [matches]

        text = ""
        while self.has_text:
            if self.startswith(*matches):
                break

            text += self.pop()

        return text

    def take_while(self, *matches: str) -> str:
        text = ""
        while self.has_text:
            if not self.startswith(*matches):
                break

            text += self.pop()

        return text

    def error(self, msg: str) -> Exception:
        return ParserException.create_from_parser(
            msg=msg,
            file=self.file,
            source=self.original,
            position=self.position,
        )



import ast
import typing_extensions as t
from .. import ast_utils
from . import tags, template_ast
from .scanner import Scanner

TagScanner = Scanner[tags.Tag]
T = t.TypeVar("T", bound=tags.Tag)
ParseRules: t.TypeAlias = t.List[
    t.Tuple[
        t.Type[T],
        t.Callable[
            [TagScanner, T],
            t.Optional[template_ast.Node],
        ],
    ]
]


def parse_tags_to_template_ast(
    tags: t.Sequence[tags.Tag],
) -> template_ast.TemplateBlock:
    scanner = Scanner(tags)
    body = take_tags_until(scanner)
    return body


def next_tag(scanner: TagScanner) -> t.Optional[template_ast.Node]:
    tag = scanner.pop()
    parse_rules: ParseRules = [
        (tags.IfStartTag, take_if),
        (tags.ForStartTag, take_for),
        (tags.SlotStartTag, take_slot),
        (tags.BlockStartTag, take_block),
    ]

    for rule_tag, func in parse_rules:
        if isinstance(tag, tags.PragmaTag):
            return None
        if isinstance(tag, template_ast.SingleTagNode):
            return tag
        if isinstance(tag, rule_tag):
            return func(scanner, tag)

    if isinstance(tag, tags.SlotEndTag):
        raise ValueError("Slot not opened")
    elif isinstance(tag, tags.BlockEndTag):
        raise ValueError("Block not opened")
    elif isinstance(tag, (tags.ElIfTag, tags.ElseTag, tags.IfEndTag)):
        raise ValueError("If statement not openned")
    elif isinstance(tag, tags.ForEndTag):
        raise ValueError("For loop not opened")
    else:
        raise ValueError("Unknown Tag")


def take_tags_until(
    scanner: TagScanner,
    stop_tags: t.List[t.Type[tags.CompositeTag]] = [],
) -> t.List[template_ast.Node]:
    tags = []
    while scanner.has_tokens:
        if scanner.is_next(*stop_tags):
            break

        tag = next_tag(scanner)
        if tag:
            tags.append(tag)

    return tags


def take_if(scanner: TagScanner, tag: tags.IfStartTag) -> template_ast.Node:
    if_block: t.List[template_ast.Node] = take_tags_until(
        scanner, stop_tags=[tags.ElIfTag, tags.ElseTag, tags.IfEndTag]
    )

    elif_blocks: t.List[t.Tuple[ast.expr, template_ast.TemplateBlock]] = []
    while scanner.is_next(tags.ElIfTag):
        elif_token = scanner.expect(tags.ElIfTag)
        block = take_tags_until(
            scanner, stop_tags=[tags.ElIfTag, tags.ElseTag, tags.IfEndTag]
        )
        elif_blocks.append((elif_token.condition, block))

    if scanner.accept(tags.ElseTag):
        else_block = take_tags_until(scanner, stop_tags=[tags.IfEndTag])
    else:
        else_block = None

    scanner.expect(tags.IfEndTag)
    return template_ast.IfNode(
        condition=tag.condition,
        if_block=if_block,
        elif_blocks=elif_blocks,
        else_block=else_block,
    )


def take_for(
    scanner: TagScanner, tag: tags.ForStartTag
) -> t.Union[template_ast.Node, None]:
    block = take_tags_until(scanner, stop_tags=[tags.ForEndTag])
    scanner.expect(tags.ForEndTag)

    return template_ast.ForNode(
        loop_variable=ast_utils.Name(tag.loop_variable),
        iterable=tag.iterable,
        loop_block=block,
    )


def take_slot(
    scanner: TagScanner, tag: tags.SlotStartTag
) -> t.Union[template_ast.Node, None]:
    if tag.name is None:
        return template_ast.SlotNode(name=None, default=None)

    if tag.is_required:
        default_body = None
    else:
        default_body = take_tags_until(scanner=scanner, stop_tags=[tags.SlotEndTag])

        scanner.expect(tags.SlotEndTag)

    slot = template_ast.SlotNode(
        name=tag.name,
        default=default_body,
    )
    return slot


def take_block(
    scanner: TagScanner, tag: tags.BlockStartTag
) -> t.Union[template_ast.Node, None]:
    body = take_tags_until(scanner, stop_tags=[tags.BlockEndTag])
    scanner.expect(tags.BlockEndTag)

    return template_ast.BlockNode(
        name=tag.name,
        body=body,
    )



from pathlib import Path
import typing_extensions as t
from .. import errors
from ..css import extract_css_from_html
from . import htmlify, introspection, lexer, template_ast, tree
from .postprocess import postprocess
from .preprocess import preprocess_html
from .tags import parse_tokens_to_tags


def parse_template(
    name: str,
    html: str,
    file: t.Union[Path, None] = None,
) -> template_ast.Template:
    try:
        return _parse_template(
            name=name,
            html=html,
            file=file,
        )
    except errors.ParserException as e:
        raise e
    except Exception as e:
        msg = f"Failed to parse template {name}"
        if file:
            msg += f" in {file}"

        raise errors.ParserException(msg) from e


def _parse_template(
    name: str,
    html: str,
    file: t.Union[Path, None],
) -> template_ast.Template:
    # Convert tokens into constant character
    # This is to prevent HTML parsing mangling it
    tokens = lexer.to_token_stream(html, file=file)
    tags = parse_tokens_to_tags(tokens)
    html, token_lookup = htmlify.convert_tags_to_valid_html(tags)

    # Process HTML
    html, css = extract_css_from_html(html, prefix=name)
    html = preprocess_html(html)
    # CSS in minified later

    # Reconvert the HTML back into tokens
    tags = htmlify.convert_tagged_html_to_tokens(html, token_lookup)

    # Parse tokens into a body
    body = tree.parse_tags_to_template_ast(tags)
    info = introspection.create_template_info(tags, css)
    body = postprocess(body, css, info)
    return construct_template_obj(name, file, body, css, info)


def construct_template_obj(
    name: str,
    file: t.Optional[Path],
    body: template_ast.TemplateBlock,
    css: str,
    info: introspection.TemplateInfo,
) -> template_ast.Template:
    if not info.is_layout:
        return template_ast.Template(
            name=name,
            file=file,
            body=body,
            parameters=info.parameters,
            css=css,
            components_calls=info.components_calls,
            style_includes=info.style_includes,
            layout=info.layout,
            blocks=info.blocks,
            imports=info.imports,
        )
    else:
        return template_ast.LayoutTemplate(
            name=name,
            file=file,
            body=body,
            css=css,
            parameters=info.parameters,
            components_calls=info.components_calls,
            style_includes=info.style_includes,
            layout=info.layout,
            blocks=info.blocks,
            slots=info.slots,
            has_default_slot=info.has_default_slot,
            imports=info.imports,
        )



from dataclasses import dataclass, field
import typing_extensions as t
from . import scanner, tags, template_ast


@dataclass
class TemplateInfo:
    css: str
    is_layout: bool = False
    parameters: t.List[template_ast.TemplateParameter] = field(default_factory=list)
    layout: t.Union[str, None] = None

    has_default_slot: bool = False
    slots: t.List[template_ast.SlotInfo] = field(default_factory=list)

    imports: t.List[template_ast.ImportNode] = field(default_factory=list)
    components_calls: t.List[template_ast.ComponentNode] = field(default_factory=list)
    style_includes: t.Set[str] = field(default_factory=set)
    styles_set: bool = False
    blocks: t.Set[str] = field(default_factory=set)


TagScanner = scanner.Scanner[tags.Tag]
T = t.TypeVar("T", bound=tags.Tag)
ParseRules: t.TypeAlias = t.List[
    t.Tuple[
        t.Type[T],
        t.Callable[[TemplateInfo, T], t.Optional[template_ast.Node]],
    ]
]


def create_template_info(stream: t.Sequence[tags.Tag], css: str) -> TemplateInfo:
    pragma_rules: ParseRules = [
        (template_ast.ImportNode, import_pragma),
        (template_ast.StyleNode, style_pragma),
        (template_ast.ComponentNode, component_prgama),
        (tags.IncludeTag, include_pragma),
        (tags.LayoutTag, layout_pragma),
        (tags.ParameterTag, parameter_pragma),
        (tags.SlotStartTag, slot_pragma),
        (tags.BlockStartTag, block_pragma),
    ]

    ctx = TemplateInfo(css)

    for tag in stream:
        for rule_tag, func in pragma_rules:
            if isinstance(tag, rule_tag):
                func(ctx, tag)

    return ctx


def include_pragma(
    ctx: TemplateInfo, tag: tags.IncludeTag
) -> t.Optional[template_ast.Node]:
    ctx.style_includes.add(tag.template)


def style_pragma(
    ctx: TemplateInfo, tag: tags.IncludeTag
) -> t.Optional[template_ast.Node]:
    if ctx.styles_set:
        raise ValueError("Template cannot have multiple styles tags")

    ctx.styles_set = True
    return template_ast.StyleNode()


def layout_pragma(
    ctx: TemplateInfo, tag: tags.LayoutTag
) -> t.Optional[template_ast.Node]:
    if ctx.layout is not None:
        raise ValueError("Template cannot have multiple layout tags")

    ctx.layout = tag.template
    return None


def parameter_pragma(
    ctx: TemplateInfo, tag: tags.ParameterTag
) -> t.Optional[template_ast.Node]:
    ctx.parameters.append(
        template_ast.TemplateParameter(
            name=tag.name,
            default=tag.default,
            type=tag.type,
        )
    )
    return None


def component_prgama(
    ctx: TemplateInfo, tag: template_ast.ComponentNode
) -> t.Optional[template_ast.Node]:
    call = template_ast.ComponentNode(
        component_name=tag.component_name, keywords=tag.keywords
    )
    ctx.components_calls.append(call)
    return call


def slot_pragma(ctx: TemplateInfo, token: tags.SlotStartTag) -> None:
    ctx.is_layout = True

    if token.name is None:
        if ctx.has_default_slot:
            raise ValueError("Template cannot have multiple default slots")

        ctx.has_default_slot = True

    slot = template_ast.SlotInfo(
        name=token.name,
        is_required=token.is_required,
    )
    ctx.slots.append(slot)


def block_pragma(ctx: TemplateInfo, token: tags.SlotStartTag) -> None:
    if token.name is not None:
        ctx.blocks.add(token.name)


def import_pragma(ctx: TemplateInfo, token: template_ast.ImportNode) -> None:
    ctx.imports.append(token)



from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import typing_extensions as t


class BuildError(Exception):
    pass


class ParsingWarning(Warning):
    pass


class InvalidTemplate(BuildError):
    @classmethod
    def create(
        cls,
        msg: str,
        name: str,
        file: t.Union[Path, None],
    ) -> t.Self:
        if file:
            return cls(f"{msg} in {name} \n{file.absolute()}")
        else:
            return cls(f"{msg} in {name}")


class ParserException(BuildError):
    error_info: t.Optional[ErrorInfo] = None

    @classmethod
    def create(
        cls,
        msg: str,
        file: t.Union[Path, None],
    ) -> t.Self:
        if file:
            msg = f"{msg} in {file.name}"

        return cls(msg)

    @classmethod
    def create_from_parser(
        cls,
        msg: str,
        source: str,
        position: int,
        file: t.Union[Path, None],
    ) -> t.Self:
        line_no, offset = calculate_line_and_offset(source, position)
        err_info = ErrorInfo(
            msg=msg,
            file=file,
            source=source,
            line_no=line_no,
            offset=offset,
        )
        msg = "\n".join(
            (
                err_info.create_error_message(),
                err_info.prev_line or "",
                err_info.err_line,
                f"{' ' * err_info.offset}^",
                err_info.next_line or "",
            )
        )

        err = cls(msg)
        err.error_info = err_info
        return err


@dataclass
class ErrorInfo:
    file: t.Optional[Path]
    msg: str
    source: str
    line_no: int
    offset: int

    @property
    def lines(self) -> list[str]:
        return self.source.split("\n")

    @property
    def err_line(self) -> str:
        line_index = self.line_no - 1
        return self.lines[line_index]

    @property
    def prev_line(self) -> t.Union[str, None]:
        line_index = self.line_no - 1
        try:
            return self.lines[line_index - 1]
        except KeyError:
            return None

    @property
    def next_line(self) -> t.Union[str, None]:
        line_index = self.line_no - 1
        try:
            return self.lines[line_index + 1]
        except KeyError:
            return None

    def create_error_message(self) -> str:
        if self.file:
            return (
                f"{self.msg} in {self.file.name} on line {self.line_no}, offset {self.offset} \n"
                f"{self.file.absolute()}:{self.line_no}:{self.offset}"
            )
        else:
            return f"{self.msg} on line {self.line_no}, offset {self.offset}"


def calculate_line_and_offset(source: str, position: int) -> t.Tuple[int, int]:
    line_index = source[:position].count("\n")
    line_no = line_index + 1

    line_start = source.rfind("\n", 0, position) + 1  # + 1 for removed newline
    offset = position - line_start
    return line_no, offset



from __future__ import annotations
import ast
import typing_extensions as t
from .. import ast_utils
from ..parser import template_ast as tags
from .constants import (
    CSS_VARIABLE,
    KWARGS_VARIABLE,
    WITH_STYLES_PARAMETER,
    create_escape_call,
    slot_parameter,
    slot_variable_name,
)

if t.TYPE_CHECKING:
    from .builder import BuildContext


T = t.TypeVar("T", bound=tags.Node, infer_variance=True)


class Rule(t.Generic[T]):
    tag: t.Type[T]

    @staticmethod
    def construct(ctx: BuildContext, tag: T):
        raise NotImplementedError


class HtmlRule(Rule[tags.HtmlNode]):
    tag = tags.HtmlNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.HtmlNode):
        ctx.add_expr(ast_utils.Constant(tag.html))


class ExprRule(Rule[tags.ExprNode]):
    tag = tags.ExprNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.ExprNode):
        ctx.add_expr(create_escape_call(value=tag.value))


class RawExprRule(Rule[tags.RawExprNode]):
    tag = tags.RawExprNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.RawExprNode):
        ctx.add_expr(tag.value)


class AssignmentRule(Rule[tags.AssignmentNode]):
    tag = tags.AssignmentNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.AssignmentNode):
        assign = ast_utils.Assign(
            target=tag.target,
            value=tag.value,
        )
        ctx.body.append(assign)


class ImportRule(Rule[tags.ImportNode]):
    tag = tags.ImportNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.ImportNode):
        template_func = ast_utils.Index(
            ast_utils.Name("_name_lookup"),
            ast_utils.Constant(tag.name),
        )
        assign = ast_utils.Assign(
            target=tag.target,
            value=template_func,
        )
        ctx.body.append(assign)


class IfRule(Rule[tags.IfNode]):
    tag = tags.IfNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.IfNode):
        ctx.ensure_output_assigned()

        empty_body = [ast_utils.create_stmt("...", ast.Expr)]
        if_body = ctx.create_block(tag.if_block) or empty_body

        elif_blocks: t.List[t.Tuple[ast.expr, t.Sequence[ast.stmt]]] = []
        for condition, elif_block in tag.elif_blocks:
            elif_body = ctx.create_block(elif_block) or empty_body
            elif_blocks.append((condition, elif_body))

        if tag.else_block:
            else_body = ctx.create_block(tag.else_block) or empty_body
        else:
            else_body = None

        ctx.body.append(
            ast_utils.If(
                condition=tag.condition,
                if_body=if_body,
                elif_blocks=elif_blocks,
                else_body=else_body,
            )
        )


class ForRule(Rule[tags.ForNode]):
    tag = tags.ForNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.ForNode):
        ctx.ensure_output_assigned()

        for_body = ctx.create_block(tag.loop_block)
        if len(for_body) > 0:
            ctx.body.append(
                ast.For(
                    target=tag.loop_variable,
                    iter=tag.iterable,
                    body=for_body,
                    orelse=[],
                )
            )


class BlockRule(Rule[tags.BlockNode]):
    tag = tags.BlockNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.BlockNode):
        ctx.body.extend(
            ctx.create_variable(
                name=slot_variable_name(tag.name),
                tags=tag.body,
            )
        )


class SlotRule(Rule[tags.SlotNode]):
    tag = tags.SlotNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.SlotNode):
        slot_param = ast_utils.Name(slot_parameter(tag.name))

        if tag.default is None:
            ctx.add_expr(slot_param)
            return

        if_stmt = ast_utils.If(
            condition=ast_utils.Is(slot_param, ast_utils.None_),
            if_body=ctx.create_variable(
                name=slot_param,
                tags=tag.default,
            ),
        )
        ctx.body.append(if_stmt)
        ctx.add_expr(slot_param)


class StyleRule(Rule[tags.StyleNode]):
    tag = tags.StyleNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.StyleNode):
        if ctx.uses_layout:
            return  # Styles are placed in the layout template
        if ctx.css is None and not ctx.is_layout:
            return

        ctx.ensure_output_assigned()
        if not ctx.is_layout:
            condition = ast_utils.Name(WITH_STYLES_PARAMETER)
            value = ast_utils.Constant(f"<style>{ctx.css}</style>")
        else:
            condition = ast_utils.And(
                ast_utils.Name(WITH_STYLES_PARAMETER),
                ast_utils.Name(CSS_VARIABLE),
            )
            value = ast_utils.FormatString(
                ast_utils.Constant("<style>"),
                ast_utils.Name(CSS_VARIABLE),
                ast_utils.Constant("</style>"),
            )

        ctx.body.append(
            ast_utils.If(
                condition=condition,
                if_body=ctx.output_variable.create_add(value),
            )
        )


class ComponentRule(Rule[tags.ComponentNode]):
    tag = tags.ComponentNode

    @staticmethod
    def construct(ctx: BuildContext, tag: tags.ComponentNode):
        func = ast_utils.Name(tag.component_name)
        keywords = tag.keywords.copy()

        keywords[WITH_STYLES_PARAMETER] = ast_utils.Constant(False)
        func_call = ast_utils.Call(
            func=func,
            keywords=keywords,
            kwargs=ast_utils.Name(KWARGS_VARIABLE),
        )
        ctx.add_expr(func_call)


default_rules: t.List[type[Rule]] = [
    HtmlRule,
    ExprRule,
    RawExprRule,
    ComponentRule,
    AssignmentRule,
    IfRule,
    ForRule,
    StyleRule,
    BlockRule,
    SlotRule,
    ImportRule,
]



import typing_extensions as t
from ..errors import InvalidTemplate
from ..parser.template_ast import LayoutTemplate, Template
from .preprocess import calculate_dependencies


def validate_templates(templates: t.Sequence[Template]):
    template_lookup = {template.name: template for template in templates}
    for template in templates:
        _check_for_missing_layout(template, template_lookup)
        _check_for_missing_blocks(template, template_lookup)
        _check_for_non_existant_blocks(template, template_lookup)
        _check_for_duplicate_parameters(template)


def _check_for_missing_layout(template: Template, lookup: t.Dict[str, Template]):
    if template.layout is None:
        return

    if template.layout not in lookup:
        raise InvalidTemplate.create(
            msg=f"Missing layout '{template.layout}'",
            name=template.name,
            file=template.file,
        )

    if not isinstance(lookup[template.layout], LayoutTemplate):
        raise InvalidTemplate.create(
            msg=f"'{template.layout}' is not a layout",
            name=template.name,
            file=template.file,
        )


def _check_for_missing_blocks(template: Template, lookup: t.Dict[str, Template]):
    if template.layout is None:
        return

    layout = t.cast(LayoutTemplate, lookup[template.layout])
    for slot in layout.slots:
        slot_used = slot.name in template.blocks
        if slot.is_required and not slot_used:
            raise InvalidTemplate.create(
                msg=f"'{template.layout}' requires slot '{slot.name}'",
                name=template.name,
                file=template.file,
            )


def _check_for_non_existant_blocks(template: Template, lookup: t.Dict[str, Template]):
    if template.layout is None:
        return

    layout = t.cast(LayoutTemplate, lookup[template.layout])
    slot_names = [slot.name for slot in layout.slots]
    for block in template.blocks:
        if block not in slot_names:
            raise InvalidTemplate.create(
                msg=f"'{template.layout}' doesn't have slot '{block}'",
                name=template.name,
                file=template.file,
            )


def _check_for_duplicate_parameters(template: Template):
    names = [param.name for param in template.parameters]
    if len(names) != len(set(names)):
        raise InvalidTemplate.create(
            msg="Duplicate parameter names",
            name=template.name,
            file=template.file,
        )



from __future__ import annotations
import ast
from dataclasses import dataclass, field
import typing_extensions as t
from ..parser.template_ast import LayoutTemplate, Node, Template
from .accumulators import Variable

if t.TYPE_CHECKING:
    from .rules import Rule


@dataclass
class BuildContext:
    template: Template
    layout: t.Optional[LayoutTemplate]
    css: t.Optional[str]
    output_variable: Variable
    rules: list[t.Type[Rule]]
    css_variable: t.Optional[Variable] = None
    body: t.List[ast.stmt] = field(default_factory=list)

    @property
    def uses_layout(self):
        return self.layout is not None

    @property
    def is_layout(self):
        return self.template.is_layout

    def add_expr(self, expr: ast.expr):
        self.body.append(self.output_variable.create_add(expr))

    def ensure_output_assigned(self):
        if not self.output_variable.assigned:
            self.body.extend(self.output_variable.assign())

    def construct_tag(self, tag: Node):
        rules = {rule.tag: rule for rule in self.rules}
        rule = rules.get(type(tag))
        if rule is None:
            raise NotImplementedError(f"Tag {tag} not implemented")

        rule.construct(self, tag)

    def create_block(self, tags: t.Sequence[Node]) -> t.Sequence[ast.stmt]:
        ctx_block = self.create_subcontext()
        for tag in tags:
            ctx_block.construct_tag(tag)

        return ctx_block.body

    def create_variable(
        self,
        name: t.Union[str, ast.Name],
        tags: t.Sequence[Node],
    ) -> t.List[ast.stmt]:
        ctx_assign = self.create_subcontext(name)

        for tag in tags:
            ctx_assign.construct_tag(tag)

        ctx_assign.ensure_output_assigned()
        return ctx_assign.body

    def create_subcontext(self, name: t.Union[str, ast.Name, None] = None):
        if name is None:
            variable = self.output_variable
        else:
            variable = Variable(name)

        return BuildContext(
            output_variable=variable,
            template=self.template,
            layout=self.layout,
            css=self.css,
            rules=self.rules,
        )



import sys
from typing import Any

SAFE_CONVERSIONS = (int, float)

if sys.version_info <= (3, 9):
    from markupsafe import escape_silent

    def escape(value: Any) -> str:
        if type(value) in SAFE_CONVERSIONS:
            return str(value)
        else:
            return str(escape_silent(value))

else:

    def escape(value: Any) -> str:
        if type(value) in SAFE_CONVERSIONS:
            return str(value)

        return (
            str(value)
            .replace("&", "&amp;")
            .replace(">", "&gt;")
            .replace("<", "&lt;")
            .replace("'", "&#39;")
            .replace('"', "&#34;")
        )



from dataclasses import dataclass
from pathlib import Path
import typing_extensions as t
from . import build, parser, types
from .compiler.constants import NAME_LOOKUP_VARIABLE
from .enviroment import Environment, Template

BUILD_FILE = Path(__file__).parent.joinpath("generated/__components.py")


@dataclass
class TemplateInfo:
    name: str
    body: str
    file: t.Optional[Path]

    def parse(self) -> parser.Template:
        return parser.parse_template(self.name, self.body, self.file)


class Tempered:
    _templates: t.List[TemplateInfo]
    _globals: t.Dict[str, t.Any]

    def __init__(self, template_folder: t.Union[str, Path, None] = None):
        self._templates = []
        self._globals = {}
        if template_folder:
            self.add_template_folder(template_folder)

    @property
    def template_files(self) -> t.List[Path]:
        return [
            template.file for template in self._templates if template.file is not None
        ]

    def add_global(self, name: str, value: t.Any):
        self._globals[name] = value

    def add_template_folder(self, folder: t.Union[Path, str]):
        folder = Path(folder)
        for file in folder.glob("**/*.*"):
            name = str(file)[len(str(folder) + "/") :]
            body = file.read_text()
            template_info = TemplateInfo(
                name=name,
                body=body,
                file=file,
            )
            self._templates.append(template_info)

    def add_template(self, file: t.Union[Path, str]):
        file = Path(file)
        name = str(file)
        body = file.read_text()
        template_info = TemplateInfo(
            name=name,
            body=body,
            file=file,
        )
        self._templates.append(template_info)

    def add_template_from_string(self, name: str, template: str):
        template_info = TemplateInfo(
            name=name,
            body=template,
            file=None,
        )
        self._templates.append(template_info)

    def _parse_templates(self) -> t.List[parser.Template]:
        return [template.parse() for template in self._templates]

    def build_enviroment(self, generate_types: bool = True) -> Environment:
        template_objs = self._parse_templates()
        m = build.build_memory(template_objs, self._globals)
        if generate_types:
            types.build_types(template_objs)
        else:
            types.clear_types()

        templates: t.Dict[str, Template] = {}
        for template in template_objs:
            func = getattr(m, NAME_LOOKUP_VARIABLE)[template.name]
            templates[template.name] = Template(func, template)  # type: ignore

        return Environment(templates=templates, globals=self._globals)



from zlib import crc32
import bs4
import tinycss2
from tinycss2.ast import HashToken, IdentToken, LiteralToken, QualifiedRule
import typing_extensions as t

# QualifiedRule = #a .b {}


# https://geoffrich.net/posts/svelte-scoped-styles/
def apply_scope_to_css(css: str, scope: str) -> str:
    css_tokens = tinycss2.parse_stylesheet(css)
    for token in list(css_tokens):
        if isinstance(token, QualifiedRule):
            add_scope_to_rule(token, scope)

    return tinycss2.serialize(css_tokens)


# QualifiedRule.prelude = #a .b
def add_scope_to_rule(rule: QualifiedRule, scope: str):
    checked_tokens = []

    # Find first scopable rule
    def insert_scope(rule: QualifiedRule, scope: str, position: int):
        scope_ident = IdentToken(0, 0, scope)
        class_token = LiteralToken(0, 0, ".")
        rule.prelude.insert(position + 1, scope_ident)
        rule.prelude.insert(position + 1, class_token)
        checked_tokens.extend((scope_ident, class_token))

    for token in list(rule.prelude):
        checked_tokens.append(token)

        if isinstance(token, (IdentToken, HashToken)):
            insert_scope(rule, scope, rule.prelude.index(token))
            break

    # Find last scopable rule
    for token in reversed(list(rule.prelude)):
        if token in checked_tokens:
            break  # We met in the middle

        if isinstance(token, (IdentToken, HashToken)):
            insert_scope(rule, scope, rule.prelude.index(token))
            break


counter = 0


def generate_scope_id(prefix: t.Union[str, None] = None) -> str:
    global counter
    counter += 1
    id = str(counter).encode()
    hash = hex(crc32(id))[2:]

    if prefix is None:
        return hash
    else:
        prefix = prefix.replace("-", "_").lower()
        return f"{prefix}-{hash[:6]}"


def apply_scope_to_soup(soup: bs4.BeautifulSoup, scope_id: str):
    for tag in soup.find_all():
        if is_tag_in_head(tag):
            continue

        classes = tag.attrs.get("class", "")

        if isinstance(classes, list):
            classes = " ".join(classes)

        tag.attrs["class"] = classes + " " + scope_id


def is_tag_in_head(tag: bs4.Tag) -> bool:
    if tag.parent is None:
        return False

    if tag.name == "head" or tag.parent.name == "head":
        return True

    return is_tag_in_head(tag.parent)



import random
import typing_extensions as t
from . import tags, template_ast
from .scanner import TextScanner

TOKEN_LENGTH = 16
TOKEN_REGEX = rf"\ TEMPERED_[A-Z0-9]{{{TOKEN_LENGTH}}}"


def generate_token_id() -> str:
    def randbytes(length: int):
        return bytes(random.randint(0, 255) for _ in range(length))

    id = randbytes(TOKEN_LENGTH // 2).hex().upper()
    return f"TEMPERED {id}"
    # Must have space before to force it to be quoted


def convert_tags_to_valid_html(
    _tags: t.Sequence[tags.Tag],
) -> t.Tuple[str, t.Dict[str, tags.Tag]]:
    html = ""
    token_lookup: t.Dict[str, tags.Tag] = {}
    for tag in _tags:
        if isinstance(tag, template_ast.HtmlNode):
            html += tag.html
        else:
            token_id = generate_token_id()
            token_lookup[token_id] = tag
            html += token_id

    return html, token_lookup


def convert_tagged_html_to_tokens(
    html: str, token_lookup: t.Dict[str, tags.Tag]
) -> t.Sequence[tags.Tag]:
    TOKEN_ID_LENGTH = len(generate_token_id())

    scanner = TextScanner(html)
    tokens_: t.List[tags.Tag] = []
    while scanner.has_text:
        known_keys = list(token_lookup.keys())
        text = scanner.take_until(known_keys)
        if len(text) > 0:
            tokens_.append(template_ast.HtmlNode(text))

        if not scanner.has_text:
            break

        token_id = scanner.pop_many(TOKEN_ID_LENGTH)
        tokens_.append(token_lookup.pop(token_id))

    return tokens_



import ast
import typing_extensions as t
from .. import ast_utils


class Variable:
    name: ast.Name
    assigned: bool = False

    def __init__(self, name: t.Union[str, ast.Name]):
        if isinstance(name, str):
            name = ast_utils.Name(name)
        self.name = name

    def create_add(self, value: ast.expr) -> ast.stmt:
        if self.assigned:
            return ast_utils.AddAssign(target=self.name, value=value)
        else:
            self.assigned = True
            return ast_utils.Assign(target=self.name, value=value)

    def assign(self, value: ast.expr = ast_utils.EmptyStr) -> t.List[ast.stmt]:
        self.assigned = True
        return [ast_utils.Assign(target=self.name, value=value)]



import string
from dataclasses import dataclass
from pathlib import Path
import typing_extensions as t
from .scanner import TextScanner

CONTROL_ESCAPE = r"\{"
EXPR_START = "{{"
EXPR_END = "}}"
STATEMENT_START = "{%"
STATEMENT_END = "%}"
COMPONENT_START = "{<"
COMPONENT_END = ">}"
COMPONENT_END_ALTERNATIVE = "/>}"
IDENT_LETTERS = (
    "abcdefghijklmnopqrstuvwxyz" "ABCDEFGHIJKLMNOPQRSTUVWXYZ" "0123456789" "_"
)
WHITESPACE = " \t\n\r\v\f"
IDENT_START = string.ascii_letters + "_"
IDENT_CONTINUE = IDENT_START + string.digits


class Token:
    pass


@dataclass
class StatementStartToken(Token):
    pass


@dataclass
class StatementEndToken(Token):
    pass


@dataclass
class ExprStartToken(Token):
    pass


@dataclass
class ComponentStartToken(Token):
    pass


@dataclass
class ComponentEndToken(Token):
    pass


@dataclass
class ExprEndToken(Token):
    pass


@dataclass
class HtmlToken(Token):
    html: str


@dataclass
class PythonExprToken(Token):
    expr: str


@dataclass
class PythonStmtToken(Token):
    stmt: str


@dataclass
class StringToken(Token):
    string: str


@dataclass
class IdentToken(Token):
    name: str


@dataclass
class KeywordToken(Token):
    keyword: str


def to_token_stream(html: str, file: t.Union[Path, None] = None) -> t.Sequence[Token]:
    scanner = TextScanner(html, file)
    _tokens: t.List[Token] = []
    while scanner.has_text:
        _tokens.extend(next_token(scanner))

    return _tokens


def next_token(scanner: TextScanner) -> t.Iterable[Token]:
    if scanner.startswith(STATEMENT_START):
        yield from next_statement_token(scanner)
    elif scanner.startswith(EXPR_START):
        yield from next_expr_token(scanner)
    elif scanner.startswith(COMPONENT_START):
        yield from next_component_token(scanner)
    else:
        yield from next_html_token(scanner)


def next_expr_token(scanner: TextScanner) -> t.Iterable[Token]:
    yield take_token(scanner, EXPR_START, ExprStartToken)
    take_whitespace(scanner)
    yield take_python_expr(scanner, EXPR_END)
    yield take_token(scanner, EXPR_END, ExprEndToken)


def next_component_token(scanner: TextScanner) -> t.Iterable[Token]:
    yield take_token(scanner, [COMPONENT_START], ComponentStartToken)
    yield take_python_expr(scanner, COMPONENT_END, COMPONENT_END_ALTERNATIVE)
    yield take_token(
        scanner, [COMPONENT_END, COMPONENT_END_ALTERNATIVE], ComponentEndToken
    )


def next_html_token(scanner: TextScanner) -> t.Iterable[Token]:
    body = ""
    while scanner.has_text:
        if scanner.startswith(EXPR_START, STATEMENT_START, COMPONENT_START):
            break
        elif scanner.accept(CONTROL_ESCAPE):
            body += "{"
        else:
            body += scanner.pop()

    yield HtmlToken(body)


def next_statement_token(scanner: TextScanner) -> t.Iterable[Token]:
    yield take_token(scanner, STATEMENT_START, StatementStartToken)
    take_whitespace(scanner)

    keyword = take_ident(scanner)
    yield KeywordToken(keyword)
    take_whitespace(scanner)

    if keyword == "if":  # {% if expr %}
        yield take_python_expr(scanner, STATEMENT_END)
    elif keyword == "elif":  # {% elif expr %}
        yield take_python_expr(scanner, STATEMENT_END)
    elif keyword == "html":  # {% rahtmlw expr %}
        yield take_python_expr(scanner, STATEMENT_END)
    elif keyword == "set":  # {% set name = expr %}
        yield take_ident_token(scanner)
        take_whitespace(scanner)
        yield take_keyword(scanner, "=")
        take_whitespace(scanner)
        yield take_python_expr(scanner, STATEMENT_END)
    elif keyword == "param":  # {% param expr %}
        yield take_python_stmt(scanner, STATEMENT_END)
    elif keyword == "layout":  # {% layout string %}
        yield take_string_token(scanner)
    elif keyword == "include":  # {% include string %}
        yield take_string_token(scanner)
    elif keyword == "block":  # {% block ident %}
        yield take_ident_token(scanner)
    elif keyword == "import":  # {% import Name from "" %}
        yield take_ident_token(scanner)
        yield take_keyword(scanner, "from")
        yield take_string_token(scanner)
    elif keyword == "for":  # {% for ident in expr %}
        yield take_ident_token(scanner)
        yield take_keyword(scanner, "in")
        yield take_python_expr(scanner, STATEMENT_END)
    elif keyword == "slot":  # {% slot name? required? %}
        yield from slot_token(scanner)
    elif keyword in {"else", "endif", "endfor", "styles", "endslot", "endblock"}:
        pass
    else:
        raise scanner.error(f'Unknown Statement "{keyword}"')

    take_whitespace(scanner)
    yield take_token(scanner, STATEMENT_END, StatementEndToken)


def slot_token(scanner: TextScanner) -> t.Iterable[Token]:
    if scanner.startswith(*IDENT_START):
        yield take_ident_token(scanner)

    if scanner.accept("required"):
        yield KeywordToken("required")

    take_whitespace(scanner)


def take_keyword(scanner: TextScanner, keyword: str) -> Token:
    scanner.expect(keyword)
    take_whitespace(scanner)
    return KeywordToken(keyword)


def take_whitespace(scanner: TextScanner):
    WHITESPACE_CHARS = list(string.whitespace)
    scanner.take_while(*WHITESPACE_CHARS)


def take_python_expr(scanner: TextScanner, *stop_tokens: str) -> Token:
    expr = scanner.take_until(stop_tokens).rstrip()
    take_whitespace(scanner)
    return PythonExprToken(expr)


def take_python_stmt(scanner: TextScanner, *stop_tokens: str) -> Token:
    expr = scanner.take_until(stop_tokens).rstrip()
    take_whitespace(scanner)
    return PythonStmtToken(expr)


def take_token(
    scanner: TextScanner,
    matches: t.Union[str, t.Sequence[str]],
    token: t.Type[Token],
) -> Token:
    if isinstance(matches, str):
        matches = [matches]

    scanner.expect(*matches)
    return token()


def take_ident_token(scanner: TextScanner) -> Token:
    ident = take_ident(scanner)
    take_whitespace(scanner)
    return IdentToken(ident)


def take_ident(scanner: TextScanner) -> str:
    ident = scanner.take(*IDENT_START)
    ident += scanner.take_while(*IDENT_CONTINUE)
    return ident


def take_string_token(scanner: TextScanner) -> Token:
    TERMINATORS = ("'", '"')

    terminator = scanner.take(*TERMINATORS)
    string = ""

    while scanner.has_text:
        if scanner.startswith("\n"):
            raise scanner.error("Newline in the middle of the string")

        if scanner.startswith("\\\\"):
            string += "\\"
        elif scanner.startswith(terminator):
            scanner.expect(terminator)
            take_whitespace(scanner)
            return StringToken(string)
        else:
            string += scanner.pop()

    raise scanner.error("String was not closed")



"Generate native python functions from HTML templates"
__version__ = "0.9.0"
from . import errors as errors
from .enviroment import Environment as Environment
from .enviroment import Template as Template
from .errors import BuildError, InvalidTemplate, ParserException
from .tempered import Tempered as Tempered

__all__ = [
    "Tempered",
    "Environment",
    "Template",
    "errors",
    "ParserException",
    "InvalidTemplate",
    "BuildError",
]



from typing import Sequence
import typing_extensions as t
from ..parser import Template


def calculate_required_css(
    template: Template,
    lookup: t.Dict[str, Template],
) -> str:
    components_used = calculate_dependencies(template, lookup)
    css_fragments = [lookup[comp].css for comp in components_used]
    return " ".join(css_fragments)


component_cache: t.Dict[str, t.Set[str]] = {}


def calculate_dependencies(
    template: Template,
    lookup: t.Dict[str, Template],
) -> Sequence[str]:
    dependencies = _recursively_calculate_dependencies(template, lookup)
    component_cache.clear()
    return dependencies


def _recursively_calculate_dependencies(
    template: Template,
    lookup: t.Dict[str, Template],
) -> Sequence[str]:
    if template.name in component_cache:
        return list(component_cache[template.name])

    components_used: t.Set[str] = set()
    components_used.add(template.name)

    required_names = [import_.name for import_ in list(template.imports)]
    if template.layout:
        required_names.append(template.layout)

    for name in list(required_names):
        child = lookup[name]
        child_children = calculate_dependencies(child, lookup)
        components_used.update(child_children)

    component_cache[template.name] = components_used
    return list(components_used)



import ast
import string
from copy import deepcopy
from pathlib import Path
import typing_extensions as t
from . import ast_utils, parser

ORIGINAL_FILE = Path(__file__).parent.joinpath("enviroment.py")
TYPES_FILE = ORIGINAL_FILE.with_suffix(".pyi")
T_OVERLOAD = ast_utils.create_expr("t.overload")


def clear_types() -> None:
    if TYPES_FILE.exists():
        TYPES_FILE.unlink()


def build_types(templates: t.List[parser.Template]) -> None:
    with open(ORIGINAL_FILE) as f:
        source = f.read()

    body = ast_utils.parse(source)

    template_classes = [create_template_class(template, body) for template in templates]
    body.extend(template_classes)

    get_template_overloads = [
        create_get_template_overload(template) for template in templates
    ]

    enviroment = find_class(body, "Environment")
    override_enviroment(enviroment, get_template_overloads)

    source = ast_utils.unparse(ast_utils.Module(body))
    TYPES_FILE.write_text(source)


def find_class(body: t.List[ast.stmt], name: str) -> ast.ClassDef:
    for node in body:
        if isinstance(node, ast.ClassDef):
            if node.name == name:
                return node

    raise RuntimeError("Could not find Environment class")


def create_get_template_overload(template: parser.Template) -> ast.FunctionDef:
    """Creates
    ```python
    def get_template(self, name: Literal["foo.html"]) -> _FooTemplate:
        ...
    ```
    """
    class_name = create_template_class_name(template)

    func_def = ast_utils.create_stmt(
        "def get_template(self, name: str) -> t.Optional[Template]: ...",
        ast.FunctionDef,
    )
    name_arg = func_def.args.args[1]
    name_arg.annotation = create_t_literal(template.name)
    func_def.returns = ast_utils.Name(class_name)
    func_def.decorator_list = [T_OVERLOAD]
    return func_def


def create_template_class(
    template: parser.Template, body: t.List[ast.stmt]
) -> ast.ClassDef:
    """Creates an object similar to this
    ```python
    class _ExampleTemplate
        def render(self, *, foo: str, bar: str, context: t.Any)
    ```"""

    class_def = deepcopy(find_class(body, "Template"))
    class_def.name = create_template_class_name(template)
    template_params = [
        ast_utils.Arg(param.name, param.type) for param in template.parameters
    ]
    param_defaults = [param.default for param in template.parameters]
    render_func = t.cast(ast.FunctionDef, class_def.body[0])
    render_func.args = ast_utils.Arguments(
        args=[ast_utils.Arg("self")],
        kwonlyargs=template_params,
        kw_defaults=param_defaults,
        kwarg=ast_utils.Arg("context", ast_utils.Name("t.Any")),
    )
    return class_def


def override_enviroment(enviroment: ast.ClassDef, overloads: t.List[ast.FunctionDef]):
    GET_TEMPLATE_NONE_OVERLOAD = ast_utils.create_stmt(
        "def get_template(self, name: str) -> None: ...",
        ast.FunctionDef,
    )
    GET_TEMPLATE_NONE_OVERLOAD.decorator_list = [T_OVERLOAD]

    for overload in overloads:
        enviroment.body.insert(0, overload)

    enviroment.body.insert(0, GET_TEMPLATE_NONE_OVERLOAD)


def create_t_literal(value: str) -> ast.expr:
    return ast_utils.Index(
        ast_utils.create_expr("t.Literal"),
        ast_utils.Constant(value),
    )


def create_template_class_name(template: parser.Template) -> str:
    IDENT_CHARS = string.ascii_letters + string.digits + "_"
    template_ident = "".join(ch for ch in template.name if ch in IDENT_CHARS)
    return f"_{template_ident}Template"



from string import whitespace
from minify_html import minify  # type: ignore


def preprocess_html(html: str) -> str:
    html = html.strip(whitespace)
    html = minify_html(html)
    return html


def minify_html(html: str) -> str:
    return minify(
        html,
        minify_js=True,
        do_not_minify_doctype=True,
        ensure_spec_compliant_unquoted_attribute_values=True,
        keep_spaces_between_attributes=True,
    )



import textwrap
import warnings
import typing_extensions as t
from .. import errors

try:
    import sass
except ImportError:
    sass = None


def transform_sass(css: str, lang: t.Literal["sass", "scss"]) -> str:
    if sass is None:
        warnings.warn(
            message="sass support not installed, run `pip install tempered[sass]`",
            category=errors.ParsingWarning,
        )
        return ""

    if lang == "sass":
        css = textwrap.dedent(css)
        return sass.compile(
            string=css,
            output_style="compressed",
            indented=True,  # sass rules
        )
    else:
        return sass.compile(
            string=css,
            output_style="compressed",
            indented=False,  # scss rules
        )



import ast
import typing_extensions as t
from .. import ast_utils
from ..parser import LayoutTemplate, Template, TemplateParameter
from .accumulators import Variable
from .builder import BuildContext
from .constants import (
    CSS_VARIABLE,
    KWARGS_VARIABLE,
    OUTPUT_VARIABLE,
    WITH_STYLES_PARAMETER,
    component_func_name,
    create_layout_call,
    layout_func_name,
    slot_parameter,
)
from .resolve import create_resolve_for_unknown_variables
from .rules import default_rules


def create_template_function(
    template: Template,
    layout: t.Union[LayoutTemplate, None],
    css: str,
) -> ast.FunctionDef:
    arguements = [*template.parameters]
    arguements.append(
        TemplateParameter(
            name=WITH_STYLES_PARAMETER,
            type=ast_utils.Name("bool"),
            default=ast_utils.Constant(True),
        )
    )

    if not isinstance(template, LayoutTemplate):
        function_name = component_func_name(template.name)
    else:
        function_name = layout_func_name(template.name)
        arguements.append(
            TemplateParameter(
                name=CSS_VARIABLE,
                type=ast_utils.Str,
            )
        )

        for slot in template.slots:
            if slot.is_required:
                type = ast_utils.Str
                default = None
            else:
                type = ast_utils.Union(ast_utils.Str, ast_utils.None_)
                default = ast_utils.Constant(None)

            arguements.append(
                TemplateParameter(
                    name=slot_parameter(slot.name),
                    type=type,
                    default=default,
                )
            )

    ctx = BuildContext(
        template=template,
        output_variable=Variable(OUTPUT_VARIABLE),
        layout=layout,
        css=css if len(css) > 0 else None,
        rules=default_rules,
    )
    func = ast_utils.Function(
        name=function_name,
        args=construct_arguments(arguements),
        body=construct_body(ctx),
        returns=ast_utils.Name("str"),
    )

    parameter_names = [param.name for param in ctx.template.parameters]
    component_names = [comp.component_name for comp in ctx.template.components_calls]
    create_resolve_for_unknown_variables(
        func.body, [*parameter_names, *component_names]
    )
    ast.fix_missing_locations(func)
    return func


def construct_arguments(arguments: t.List[TemplateParameter]) -> ast.arguments:
    args = []
    defaults = []
    for arguement in arguments:
        args.append(
            ast.arg(
                arg=arguement.name,
                annotation=arguement.type,
            )
        )
        defaults.append(arguement.default)

    return ast_utils.Arguments(
        kwonlyargs=args,
        kw_defaults=defaults,
        kwarg=ast_utils.Arg(KWARGS_VARIABLE, ast_utils.create_expr("t.Any")),
    )


def construct_body(ctx: BuildContext) -> t.Sequence[ast.AST]:
    statements: t.List[ast.AST] = []
    if ctx.is_layout or ctx.uses_layout:
        statements.extend(create_style_contant(ctx))

    for tag in ctx.template.body:
        ctx.construct_tag(tag)

    ctx.ensure_output_assigned()
    output_value = ctx.output_variable.name

    statements.extend(ctx.body)

    if ctx.layout:
        output_value = create_layout_call(
            layout_name=ctx.layout.name,
            default_slot=output_value,
            css=ast_utils.Name(CSS_VARIABLE),
            has_default_slot=ctx.layout.has_default_slot,
            blocks=ctx.template.blocks,
        )

    statements.append(ast_utils.Return(output_value))
    return statements


def create_style_contant(ctx: BuildContext) -> t.List[ast.stmt]:
    if ctx.is_layout:
        return []

    if ctx.css is None and not ctx.uses_layout:
        return []

    value = ast_utils.Constant(ctx.css or "")

    ctx.css_variable = Variable(CSS_VARIABLE)
    return ctx.css_variable.assign(value)



from __future__ import annotations
import ast
from abc import ABC
from dataclasses import dataclass, field
from pathlib import Path
import typing_extensions as t


class Node(ABC):
    pass


class SingleTagNode(Node):
    pass


@dataclass
class HtmlNode(SingleTagNode):
    html: str


@dataclass
class ExprNode(SingleTagNode):
    value: ast.expr


@dataclass
class RawExprNode(SingleTagNode):
    value: ast.expr


@dataclass
class StyleNode(SingleTagNode):
    """Place styles in component here"""


@dataclass
class ComponentNode(SingleTagNode):
    # Needed to prevent CSS from being created multiple times
    # Also to prevent HTML from being escaped
    component_name: str
    keywords: t.Dict[str, ast.expr]


@dataclass
class AssignmentNode(SingleTagNode):
    target: ast.expr
    value: ast.expr


@dataclass
class ImportNode(SingleTagNode):
    target: str
    name: str


@dataclass
class IfNode(Node):
    condition: ast.expr
    if_block: TemplateBlock
    else_block: t.Optional[TemplateBlock]
    elif_blocks: t.List[t.Tuple[ast.expr, TemplateBlock]] = field(default_factory=list)


@dataclass
class ForNode(Node):
    loop_variable: ast.Name
    iterable: ast.expr
    loop_block: TemplateBlock


@dataclass
class SlotNode(Node):
    name: t.Optional[str]
    default: t.Optional[TemplateBlock]


@dataclass
class BlockNode(Node):
    name: t.Optional[str]
    body: TemplateBlock


@dataclass
class TemplateParameter:
    name: str
    type: t.Optional[ast.expr] = None
    default: t.Optional[ast.expr] = None


@dataclass
class SlotInfo:
    name: t.Optional[str]
    is_required: bool = False


@dataclass
class Template:
    name: str
    is_layout = False
    file: t.Optional[Path] = None
    parameters: t.List[TemplateParameter] = field(default_factory=list)
    imports: t.List[ImportNode] = field(default_factory=list)

    body: TemplateBlock = field(default_factory=list)
    css: str = ""
    layout: t.Union[str, None] = None
    components_calls: t.List[ComponentNode] = field(default_factory=list)
    style_includes: t.Set[str] = field(default_factory=set)

    blocks: t.Set[str] = field(default_factory=set)


@dataclass
class LayoutTemplate(Template):
    is_layout = True
    has_default_slot: bool = False
    slots: t.List[SlotInfo] = field(default_factory=list)


TemplateBlock: t.TypeAlias = t.Sequence[Node]



import warnings
import bs4
from bs4 import MarkupResemblesLocatorWarning
import typing_extensions as t
from .. import errors
from . import sass, scoped

warnings.simplefilter("ignore", MarkupResemblesLocatorWarning)


html: t.TypeAlias = str
css: t.TypeAlias = str


def extract_css_from_html(body: str, prefix: str = "tempered") -> t.Tuple[html, css]:
    soup = bs4.BeautifulSoup(body, "html.parser")
    css = ""

    style_tags = t.cast(t.List[bs4.Tag], soup.find_all("style"))
    has_styles = len(style_tags) != 0
    if not has_styles:
        return body, ""

    scope_id = scoped.generate_scope_id(prefix)
    scoped.apply_scope_to_soup(soup, scope_id)
    for tag in style_tags:
        try:
            styles = tag.text
            is_global = tag.has_attr("global")
            lang = tag.get("lang", None)
            if isinstance(lang, list):
                lang = lang[0]
            styles = transform_style_tag(styles, scope_id, is_global, lang)
        except Exception:
            warnings.warn(message="Failed to parse CSS", category=errors.ParsingWarning)
            styles = ""

        css += styles
        tag.decompose()

    html = soup.prettify(formatter="minimal")
    return html, css


def transform_style_tag(
    css: str, scope: str, is_global: bool, lang: t.Union[str, None]
) -> str:
    if lang == "scss" or lang == "sass":
        css = sass.transform_sass(css, lang)

    if is_global:
        return css
    else:
        return scoped.apply_scope_to_css(css, scope)



from importlib.util import module_from_spec, spec_from_loader
from types import ModuleType
import typing_extensions as t
from . import ast_utils, parser
from .compiler.constants import component_func_name
from .compiler.module import compile_module


def build_memory(
    templates: t.List[parser.Template],
    globals: t.Dict[str, t.Any],
) -> ModuleType:
    module_ast = compile_module(templates)

    source = ast_utils.unparse(module_ast)

    spec = spec_from_loader(name="tempered.components", loader=None)
    assert spec is not None
    module = module_from_spec(spec)
    exec(source, module.__dict__)

    # Register globals
    for name, value in globals.items():
        module.__register_global(name, value)

    return module


def build_single_template(
    template: parser.Template,
    templates: t.List[parser.Template],
    globals: t.Dict[str, t.Any],
) -> t.Callable[..., str]:
    module = build_memory([template, *templates], globals)
    func = getattr(module, component_func_name(template.name))
    return func



import ast
import sys
import typing_extensions as t
from typing_extensions import Any, Iterable, Sequence

NoneType = type(None)

LiteralType: t.TypeAlias = t.Union[
    None,
    str,
    bytes,
    bool,
    int,
    float,
    complex,
    list,
    set,
    tuple,
    dict,
]


def Constant(value: LiteralType) -> ast.expr:
    CONSTANT_TYPES = (NoneType, str, bytes, bool, int, float, complex)
    if isinstance(value, list):
        return List(value)
    elif isinstance(value, set):
        return Tuple(value)
    elif isinstance(value, tuple):
        return Set(value)
    elif isinstance(value, dict):
        return Dict(value)
    elif isinstance(value, CONSTANT_TYPES):
        return ast.Constant(value=value)
    else:
        raise ValueError(f"Cannot convert {value} to an ast constant")


EmptyStr = Constant("")
None_ = Constant(None)


def List(value: Iterable[t.Union[LiteralType, ast.expr]]) -> ast.List:
    return ast.List(elts=_IterableConstant(value))


def Tuple(value: Iterable[t.Union[LiteralType, ast.expr]]) -> ast.Tuple:
    return ast.Tuple(elts=_IterableConstant(value))


def Set(value: Iterable[t.Union[LiteralType, ast.expr]]) -> ast.Set:
    return ast.Set(elts=_IterableConstant(value))


def Dict(value: dict) -> ast.Dict:
    return ast.Dict(
        keys=_IterableConstant(value.keys()),
        values=_IterableConstant(value.values()),
    )


def _IterableConstant(
    iterable: Iterable[t.Union[LiteralType, ast.expr]]
) -> t.List[ast.expr]:
    values = []
    for item in iterable:
        if not isinstance(item, ast.AST):
            item = Constant(item)
        values.append(item)

    return values


def Call(
    func: ast.AST,
    arguments: Sequence[ast.expr] = [],
    keywords: t.Dict[str, ast.expr] = {},
    kwargs: t.Union[ast.Name, None] = None,
) -> ast.Call:
    call_keywords = [
        ast.keyword(arg=name, value=value) for name, value in keywords.items()
    ]
    if kwargs is not None:
        call_keywords.append(ast.keyword(value=kwargs, arg=None))

    return ast.Call(
        func=func,
        args=arguments,
        keywords=call_keywords,
    )


def Name(ident: str) -> ast.Name:
    return ast.Name(id=ident, ctx=ast.Load())


Str = Name("str")
Int = Name("int")
Float = Name("float")
Bool = Name("bool")


def Module(body: Sequence[ast.AST]) -> ast.Module:
    module = ast.Module(
        body=body,
        type_ignores=[],
    )
    ast.fix_missing_locations(module)
    return module


def Arg(name: str, annotation: t.Optional[ast.expr] = None) -> ast.arg:
    return ast.arg(name, annotation)


def Arguments(
    args: t.List[ast.arg] = [],
    defaults: t.List[ast.expr] = [],
    kw_defaults: t.List[t.Optional[ast.expr]] = [],
    posonlyargs: t.List[ast.arg] = [],
    kwonlyargs: t.List[ast.arg] = [],
    vararg: t.Optional[ast.arg] = None,
    kwarg: t.Optional[ast.arg] = None,
) -> ast.arguments:
    return ast.arguments(
        args=args,
        defaults=defaults,
        kw_defaults=kw_defaults,
        posonlyargs=posonlyargs,
        kwonlyargs=kwonlyargs,
        vararg=vararg,
        kwarg=kwarg,
    )


def Function(
    name: str,
    args: ast.arguments,
    body: Sequence[ast.AST],
    returns: t.Union[ast.expr, None] = None,
    decorators: Sequence[ast.expr] = [],
) -> ast.FunctionDef:
    return ast.FunctionDef(
        name=name,
        args=args,
        body=body,
        returns=returns,
        decorator_list=decorators,
        type_params=[],
    )


def ClassDef(
    name: str,
    bases: t.List[ast.expr] = [],
    body: t.List[ast.stmt] = [],
) -> ast.ClassDef:
    return ast.ClassDef(
        name=name,
        bases=bases,
        keywords=[],
        body=body,
        decorator_list=[],
        type_params=[],
    )


def AddAssign(
    target: t.Union[str, ast.Name], value: t.Union[LiteralType, ast.expr]
) -> ast.AugAssign:
    if isinstance(target, str):
        target = Name(target)

    if not isinstance(value, ast.AST):
        value = Constant(value)

    return ast.AugAssign(
        op=ast.Add(),
        target=target,
        value=value,
    )


def Assign(
    target: t.Union[str, ast.expr], value: t.Union[LiteralType, ast.AST]
) -> ast.Assign:
    if isinstance(target, str):
        target = Name(target)

    if not isinstance(value, ast.AST):
        value = Constant(value)

    return ast.Assign(
        targets=[target],
        value=value,
        type_comment=None,
    )


def Index(value: ast.expr, index: ast.expr) -> ast.Subscript:
    return ast.Subscript(
        value=value,
        slice=index,
        ctx=ast.Load(),
    )


def UnaryOp(op: ast.unaryop, arg: ast.expr) -> ast.expr:
    return ast.UnaryOp(
        op=op,
        operand=arg,
    )


def BoolOp(op: ast.boolop, *args: ast.expr) -> ast.expr:
    return ast.BoolOp(op=op, values=args)


def And(*args: ast.expr) -> ast.expr:
    return BoolOp(ast.And(), *args)


def BinOp(op: ast.operator, *args: ast.expr) -> ast.expr:
    if len(args) == 1:
        return args[0]
    else:
        return ast.BinOp(
            op=op,
            left=args[0],
            right=BinOp(op, *args[1:]),
        )


def Add(*args: ast.expr) -> ast.expr:
    return BinOp(ast.Add(), *args)


def Union(*args: ast.expr) -> ast.expr:
    return BinOp(ast.BitOr(), *args)


def Compare(op: ast.cmpop, left: ast.expr, right: ast.expr) -> ast.expr:
    return ast.Compare(
        left=left,
        ops=[op],
        comparators=[right],
    )


def Is(left: ast.expr, right: ast.expr) -> ast.expr:
    return Compare(ast.Is(), left, right)


def Equals(left: ast.expr, right: ast.expr) -> ast.expr:
    return Compare(ast.Eq(), left, right)


def Attribute(value: ast.expr, attr: str):
    return ast.Attribute(
        value=value,
        attr=attr,
        ctx=ast.Load(),
    )


def ArrayJoin(array: ast.expr) -> ast.Call:
    return ast.Call(
        func=ast.Attribute(value=ast.Constant(value=""), attr="join"),
        args=[array],
        keywords=[],
    )


def Comprehension(
    loop_var: t.Union[str, ast.Name],
    iterable: ast.expr,
) -> ast.comprehension:
    if isinstance(loop_var, str):
        loop_var = Name(loop_var)

    return ast.comprehension(
        target=loop_var,
        iter=iterable,
        ifs=[],
        is_async=0,
    )


def GeneratorExp(
    expr: ast.expr,
    loop_var: t.Union[str, ast.Name],
    iterable: ast.expr,
) -> ast.GeneratorExp:
    return ast.GeneratorExp(
        elt=expr,
        generators=[Comprehension(loop_var, iterable)],
    )


def ListComp(
    expr: ast.expr,
    loop_var: t.Union[str, ast.Name],
    iterable: ast.expr,
) -> ast.ListComp:
    return ast.ListComp(
        elt=expr,
        generators=[Comprehension(loop_var, iterable)],
    )


def Return(value: t.Union[ast.expr, None] = None) -> ast.Return:
    return ast.Return(value=value)


def FormatString(*expressions: ast.expr) -> ast.expr:
    values = []
    for expr in expressions:
        if isinstance(expr, ast.Constant) and isinstance(expr.value, str):
            values.append(expr)
        else:
            values.append(
                ast.FormattedValue(value=expr, conversion=-1, format_spec=None)
            )

    return ast.JoinedStr(values=values)


IfBody: t.TypeAlias = t.Union[t.Sequence[ast.stmt], ast.stmt]


def If(
    condition: ast.expr,
    if_body: IfBody,
    elif_blocks: Sequence[t.Tuple[ast.expr, IfBody]] = [],
    else_body: t.Union[IfBody, None] = None,
) -> ast.If:
    if isinstance(if_body, ast.AST):
        if_body = [if_body]

    def navigate_elif_chain(if_statement: ast.If) -> ast.If:
        cur_if = if_statement
        while len(cur_if.orelse) == 1 and isinstance(cur_if.orelse[0], ast.If):
            cur_if = cur_if.orelse[0]

        return cur_if

    def insert_elif(if_statement: ast.If, condition: ast.expr, body: IfBody):
        if isinstance(body, ast.AST):
            body = [body]

        bottom_if = navigate_elif_chain(if_statement)
        bottom_if.orelse = [
            ast.If(
                test=condition,
                body=body,
                orelse=[],
            )
        ]

    if_statement = ast.If(
        test=condition,
        body=if_body,
        orelse=[],
    )
    for elif_cond, elif_block in elif_blocks:
        insert_elif(if_statement, elif_cond, elif_block)

    if else_body is not None:
        bottom_if = navigate_elif_chain(if_statement)
        if isinstance(else_body, ast.stmt):
            else_body = [else_body]

        bottom_if.orelse = list(else_body)

    return if_statement


TStmt = t.TypeVar("TStmt", bound=ast.stmt)


def create_stmt(code: str, type: t.Type[TStmt]) -> TStmt:
    module = ast.parse(code)
    if len(module.body) != 1:
        raise RuntimeError("Generated code a single stmt")

    expr = module.body[0]
    if not isinstance(expr, type):
        raise RuntimeError(f"Generated code was not {type}")

    return t.cast(TStmt, expr)


TExpr = t.TypeVar("TExpr", bound=ast.expr)


def create_expr(code: str, type: t.Type[TExpr] = ast.expr) -> TExpr:
    expr = create_stmt(code, ast.Expr)
    if not isinstance(expr.value, type):
        raise RuntimeError(f"Generated code was not {type}")

    return expr.value


def print_ast(module: t.Union[t.List[ast.AST], ast.Module]):
    if isinstance(module, list):
        module = Module(module)

    print(unparse(module))


def parse(code: str) -> t.List[ast.stmt]:
    return ast.parse(code).body


def unparse(node: ast.AST):
    if sys.version_info >= (3, 9):
        return ast.unparse(node)
    else:
        import astor

        return astor.to_source(node).strip()



import typing_extensions as t
from . import template_ast
from .introspection import TemplateInfo
from .template_ast import Node, TemplateBlock


def postprocess(body: TemplateBlock, css: str, info: TemplateInfo) -> TemplateBlock:
    body = place_styles(body, css, info)
    return body


def place_styles(body: TemplateBlock, css: str, info: TemplateInfo) -> TemplateBlock:
    has_css_includes = len(info.style_includes) > 0 or len(info.components_calls) > 0
    has_css = len(css) > 0 or has_css_includes

    if info.styles_set:
        return body

    if not info.is_layout and not has_css:
        # If there is no CSS and it's not a layout
        # If there can never be additional css
        # So don't place a style tag
        return body

    return [*body, template_ast.StyleNode()]



import ast
from typing import Sequence
from .. import ast_utils
from ..css import finalise_css
from ..parser import LayoutTemplate, Template
from . import preprocess, validate
from .constants import FILE_HEADER, NAME_LOOKUP_VARIABLE
from .template import create_template_function


def compile_module(
    templates: Sequence[Template],
) -> ast.Module:
    validate.validate_templates(templates)

    lookup = {template.name: template for template in templates}
    layout_lookup = {
        template.name: template
        for template in templates
        if isinstance(template, LayoutTemplate)
    }

    functions = []
    name_func_map = {}
    for template in templates:
        if template.layout is None:
            layout = None
        else:
            layout = layout_lookup[template.layout]

        css = preprocess.calculate_required_css(template, lookup)
        css = finalise_css(css)
        func = create_template_function(template, layout, css)
        name_func_map[template.name] = ast_utils.Name(func.name)
        functions.append(func)

    body = [
        *FILE_HEADER,
        *functions,
        ast_utils.Assign(NAME_LOOKUP_VARIABLE, ast_utils.Constant(name_func_map)),
    ]
    return ast_utils.Module(body)



import tinycss2
from rcssmin import cssmin
from tinycss2.ast import AtRule, Node
import typing_extensions as t


def finalise_css(css: str) -> str:
    css = minify_css(css)
    css = place_font_imports_at_start(css)
    return css


def place_font_imports_at_start(css: str) -> str:
    rules: t.List[Node] = tinycss2.parse_stylesheet(css)

    def sort_rules(rule: Node) -> int:
        if isinstance(rule, AtRule):
            if rule.at_keyword == "charset":
                return 1
            if rule.at_keyword == "import":
                return 2

        return -1

    rules.sort(key=sort_rules, reverse=True)
    css = tinycss2.serialize(rules)
    return css


def minify_css(css: str) -> str:
    minified_css = cssmin(css)
    if isinstance(minified_css, str):
        return minified_css
    elif isinstance(minified_css, (bytes, bytearray)):
        return minified_css.decode()
    else:
        raise TypeError("Expected str or bytes")



from __future__ import annotations
import ast
from abc import ABC
from dataclasses import dataclass
import typing_extensions as t
from .. import ast_utils
from . import lexer, template_ast
from .scanner import Scanner


class CompositeTag(ABC):
    pass


class PragmaTag(ABC):
    pass


@dataclass
class IfStartTag(CompositeTag):
    condition: ast.expr


@dataclass
class ElIfTag(CompositeTag):
    condition: ast.expr


@dataclass
class ElseTag(CompositeTag):
    pass


@dataclass
class IfEndTag(CompositeTag):
    pass


@dataclass
class ForStartTag(CompositeTag):
    iterable: ast.expr
    loop_variable: str


@dataclass
class ForEndTag(CompositeTag):
    pass


@dataclass
class SlotStartTag(CompositeTag):
    name: t.Union[str, None]
    is_required: bool = False


@dataclass
class SlotEndTag(CompositeTag):
    pass


@dataclass
class BlockStartTag(CompositeTag):
    name: str


@dataclass
class BlockEndTag(CompositeTag):
    pass


@dataclass
class LayoutTag(PragmaTag):
    template: str


@dataclass
class IncludeTag(PragmaTag):
    template: str


@dataclass
class ParameterTag(PragmaTag):
    name: str
    type: t.Union[ast.expr, None] = None
    default: t.Union[ast.expr, None] = None


Tag: t.TypeAlias = t.Union[CompositeTag, PragmaTag, template_ast.SingleTagNode]
TokenScanner: t.TypeAlias = Scanner[lexer.Token]


def parse_tokens_to_tags(stream: t.Sequence[lexer.Token]) -> t.Sequence[Tag]:
    scanner = Scanner(stream)
    return list(_take_body(scanner))


def _take_body(scanner: TokenScanner) -> t.Iterable[Tag]:
    while scanner.has_tokens:
        yield _next_tag(scanner)


def _next_tag(scanner: TokenScanner) -> Tag:
    if scanner.is_next(lexer.StatementStartToken):
        return next_statement_tag(scanner)
    elif scanner.is_next(lexer.ExprStartToken):
        return _next_expr_tag(scanner)
    elif scanner.is_next(lexer.ComponentStartToken):
        return _next_component_tag(scanner)

    token = scanner.pop()
    if isinstance(token, lexer.HtmlToken):
        return template_ast.HtmlNode(token.html)
    else:
        raise ValueError(f"Unknown token {token}")


def _next_expr_tag(scanner: TokenScanner) -> Tag:
    scanner.expect(lexer.ExprStartToken)
    expr = take_python_expr(scanner)
    scanner.expect(lexer.ExprEndToken)
    return template_ast.ExprNode(expr)


def _next_component_tag(scanner: TokenScanner) -> Tag:
    scanner.expect(lexer.ComponentStartToken)
    call = take_python_expr(scanner, ast.Call)
    scanner.expect(lexer.ComponentEndToken)

    if not isinstance(call.func, ast.Name):
        raise ValueError("Component call must be a name")

    keywords = {
        keyword.arg: keyword.value
        for keyword in call.keywords
        if keyword.arg is not None
    }
    return template_ast.ComponentNode(
        component_name=call.func.id,
        keywords=keywords,
    )


def next_statement_tag(scanner: TokenScanner) -> Tag:
    scanner.expect(lexer.StatementStartToken)
    keyword = scanner.expect(lexer.KeywordToken).keyword

    statement_funcs: t.Dict[str, t.Callable[[TokenScanner], Tag]] = {
        "if": lambda _: IfStartTag(take_python_expr(scanner)),
        "elif": lambda _: ElIfTag(take_python_expr(scanner)),
        "else": lambda _: ElseTag(),
        "endif": lambda _: IfEndTag(),
        "for": _next_for_token,
        "endfor": lambda _: ForEndTag(),
        "slot": _next_slot_tag,
        "endslot": lambda _: SlotEndTag(),
        "block": lambda _: BlockStartTag(take_ident_token(scanner)),
        "endblock": lambda _: BlockEndTag(),
        "html": lambda _: template_ast.RawExprNode(take_python_expr(scanner)),
        "set": _next_assign_tag,
        "layout": lambda _: LayoutTag(take_string_token(scanner)),
        "styles": lambda _: template_ast.StyleNode(),
        "include": lambda _: IncludeTag(take_string_token(scanner)),
        "import": _next_import_tag,
        "param": _next_param_tag,
    }
    tag = statement_funcs[keyword](scanner)
    scanner.expect(lexer.StatementEndToken)
    return tag


def _next_assign_tag(scanner: TokenScanner) -> Tag:
    target = scanner.expect(lexer.IdentToken).name
    expect_keyword(scanner, "=")
    expr = take_python_expr(scanner)
    return template_ast.AssignmentNode(
        target=ast_utils.Name(target),
        value=expr,
    )


def _next_import_tag(scanner: TokenScanner) -> Tag:
    target = scanner.expect(lexer.IdentToken).name
    expect_keyword(scanner, "from")
    name = take_string_token(scanner)
    return template_ast.ImportNode(
        target=target,
        name=name,
    )


def _next_param_tag(scanner: TokenScanner) -> Tag:
    expr = take_python_stmt(scanner)

    if isinstance(expr, ast.AnnAssign) and isinstance(expr.target, ast.Name):
        # name: type | name: type = default
        return ParameterTag(
            name=expr.target.id,
            type=expr.annotation,
            default=expr.value,
        )
    elif (
        isinstance(expr, ast.Assign)
        and isinstance(expr.targets, list)
        and isinstance(expr.targets[0], ast.Name)
    ):
        # name = default
        return ParameterTag(
            name=expr.targets[0].id,
            default=expr.value,
        )
    elif isinstance(expr, ast.Expr) and isinstance(expr.value, ast.Name):
        # name
        return ParameterTag(name=expr.value.id)
    else:
        raise ValueError(f"Invalid Parameter: {expr}")


def _next_slot_tag(scanner: TokenScanner) -> Tag:
    if scanner.is_next(lexer.IdentToken):
        name = take_ident_token(scanner)
    else:
        name = None

    if scanner.is_next(lexer.KeywordToken):
        expect_keyword(scanner, "required")
        is_required = True
    else:
        is_required = False

    return SlotStartTag(name, is_required)


def _next_for_token(scanner: TokenScanner) -> Tag:
    ident = scanner.expect(lexer.IdentToken).name
    expect_keyword(scanner, "in")
    iterable = take_python_expr(scanner)
    return ForStartTag(iterable, ident)


def take_string_token(scanner: TokenScanner) -> str:
    string_token = scanner.expect(lexer.StringToken)
    return string_token.string


def take_ident_token(scanner: TokenScanner) -> str:
    ident_token = scanner.expect(lexer.IdentToken)
    return ident_token.name


TExpr = t.TypeVar("TExpr", bound=ast.expr)
TStmt = t.TypeVar("TStmt", bound=ast.stmt)


def take_python_expr(scanner: TokenScanner, type: t.Type[TExpr] = ast.expr) -> TExpr:
    token = scanner.expect(lexer.PythonExprToken)
    return ast_utils.create_expr(token.expr, type)


def take_python_stmt(scanner: TokenScanner, type: t.Type[TStmt] = ast.stmt) -> TStmt:
    token = scanner.expect(lexer.PythonStmtToken)
    return ast_utils.create_stmt(token.stmt, type)


def expect_keyword(scanner: TokenScanner, keyword: str):
    keyword_token = scanner.expect(lexer.KeywordToken)
    if keyword_token.keyword != keyword:
        raise ValueError(f"Expected keyword {keyword}")
